@@ -1,18 +1,18 @@
+import functools
+import inspect
 import os
 import re
 import sys
-import inspect
-import warnings
-import functools
 import threading
+import warnings
 from timeit import default_timer
 
-from flask import request, make_response, current_app
 from flask import Flask, Response
+from flask import request, make_response, current_app
 from flask.views import MethodViewType
-from werkzeug.serving import is_running_from_reloader
 from prometheus_client import Counter, Histogram, Gauge, Summary
-from prometheus_client.exposition import choose_encoder
+from prometheus_client.exposition import choose_formatter
+from werkzeug.serving import is_running_from_reloader
 
 if sys.version_info[0:2] >= (3, 4):
     # Python v3.4+ has a built-in has __wrapped__ attribute
@@ -102,12 +102,19 @@ class PrometheusMetrics(object):
         - Without an argument, possibly to use with the Flask `request` object
     """
 
-    def __init__(self, app, path='/metrics',
-                 export_defaults=True, defaults_prefix='flask',
-                 group_by='path', buckets=None,
+    def __init__(self, app,
+                 path='/metrics',
+                 export_defaults=True,
+                 defaults_prefix='flask',
+                 group_by='path',
+                 buckets=None,
                  default_latency_as_histogram=True,
-                 default_labels=None, response_converter=None,
-                 excluded_paths=None, registry=None, **kwargs):
+                 default_labels=None,
+                 response_converter=None,
+                 excluded_paths=None,
+                 exclude_user_defaults=True,
+                 metrics_decorator=None,
+                 registry=None, **kwargs):
         """
         Create a new Prometheus metrics export configuration.
 
@@ -130,8 +137,12 @@ class PrometheusMetrics(object):
             metrics exposed by this `PrometheusMetrics` instance
         :param response_converter: a function that converts the captured
             the produced response object to a Flask friendly representation
+        :param metrics_decorator: an optional decorator to apply to the
+            metrics endpoint, takes a function and needs to return a function
         :param excluded_paths: regular expression(s) as a string or
             a list of strings for paths to exclude from tracking
+        :param exclude_user_defaults: also apply the `excluded_paths`
+            exclusions to user-defined defaults (not only built-in ones)
         :param registry: the Prometheus Registry to use
         """
 
@@ -142,6 +153,7 @@ class PrometheusMetrics(object):
         self._default_labels = default_labels or {}
         self._default_latency_as_histogram = default_latency_as_histogram
         self._response_converter = response_converter or make_response
+        self._metrics_decorator = metrics_decorator
         self.buckets = buckets
         self.version = __version__
 
@@ -190,6 +202,8 @@ class PrometheusMetrics(object):
         else:
             self.excluded_paths = None
 
+        self.exclude_user_defaults = exclude_user_defaults
+
         if app is not None:
             self.init_app(app)
 
@@ -248,13 +262,12 @@ class PrometheusMetrics(object):
         if app is None:
             app = self.app or current_app
 
-        @app.route(path)
         @self.do_not_track()
         def prometheus_metrics():
             # import these here so they don't clash with our own multiprocess module
             from prometheus_client import multiprocess, CollectorRegistry
 
-            if 'prometheus_multiproc_dir' in os.environ:
+            if 'PROMETHEUS_MULTIPROC_DIR' in os.environ or 'prometheus_multiproc_dir' in os.environ:
                 registry = CollectorRegistry()
             else:
                 registry = self.registry
@@ -262,14 +275,21 @@ class PrometheusMetrics(object):
             if 'name[]' in request.args:
                 registry = registry.restricted_registry(request.args.getlist('name[]'))
 
-            if 'prometheus_multiproc_dir' in os.environ:
+            if 'PROMETHEUS_MULTIPROC_DIR' in os.environ or 'prometheus_multiproc_dir' in os.environ:
                 multiprocess.MultiProcessCollector(registry)
 
-            generate_latest, content_type = choose_encoder(request.headers.get("Accept"))
+            generate_latest, content_type = choose_formatter(request.headers.get("Accept"))
             headers = {'Content-Type': content_type}
             return generate_latest(registry), 200, headers
 
-    def start_http_server(self, port, host='0.0.0.0', endpoint='/metrics'):
+        # apply any user supplied decorators, like authentication
+        if self._metrics_decorator:
+            prometheus_metrics = self._metrics_decorator(prometheus_metrics)
+
+        # apply the Flask route decorator on our metrics endpoint
+        app.route(path)(prometheus_metrics)
+
+    def start_http_server(self, port, host='0.0.0.0', endpoint='/metrics', ssl=None):
         """
         Start an HTTP server for exposing the metrics.
         This will be an individual Flask application,
@@ -279,6 +299,10 @@ class PrometheusMetrics(object):
         :param host: the HTTP host to listen on (default: `0.0.0.0`)
         :param endpoint: the URL path to expose the endpoint on
             (default: `/metrics`)
+        :param ssl: enable SSL to http server
+            It expects a dict with 2 keys: `cert` and `key` with
+            certificate and key paths.
+            Default: `None`
         """
 
         if is_running_from_reloader():
@@ -288,10 +312,13 @@ class PrometheusMetrics(object):
         self.register_endpoint(endpoint, app)
 
         def run_app():
-            app.run(host=host, port=port)
+            if ssl is None:
+                app.run(host=host, port=port)
+            else:
+                app.run(host=host, port=port, ssl_context=(ssl["cert"], ssl["key"]))
 
         thread = threading.Thread(target=run_app)
-        thread.setDaemon(True)
+        thread.daemon = True
         thread.start()
 
     def export_defaults(self, buckets=None, group_by='path',
@@ -408,7 +435,7 @@ class PrometheusMetrics(object):
                 if any(pattern.match(request.path) for pattern in self.excluded_paths):
                     return response
 
-            if hasattr(request, 'prom_start_time'):
+            if hasattr(request, 'prom_start_time') and self._not_yet_handled('duration_reported'):
                 total_time = max(default_timer() - request.prom_start_time, 0)
 
                 if callable(duration_group):
@@ -425,10 +452,11 @@ class PrometheusMetrics(object):
 
                 request_duration_metric.labels(**request_duration_labels).observe(total_time)
 
-            request_total_metric.labels(
-                method=request.method, status=_to_status_code(response.status_code),
-                **labels.values_for(response)
-            ).inc()
+            if self._not_yet_handled('total_reported'):
+                request_total_metric.labels(
+                    method=request.method, status=_to_status_code(response.status_code),
+                    **labels.values_for(response)
+                ).inc()
 
             return response
 
@@ -452,7 +480,7 @@ class PrometheusMetrics(object):
                 **labels.values_for(response)
             ).inc()
 
-            if hasattr(request, 'prom_start_time'):
+            if hasattr(request, 'prom_start_time') and self._not_yet_handled('duration_reported'):
                 total_time = max(default_timer() - request.prom_start_time, 0)
 
                 request_duration_labels = {
@@ -464,10 +492,11 @@ class PrometheusMetrics(object):
 
                 request_duration_metric.labels(**request_duration_labels).observe(total_time)
 
-            request_total_metric.labels(
-                method=request.method, status=500,
-                **labels.values_for(response)
-            ).inc()
+            if self._not_yet_handled('total_reported'):
+                request_total_metric.labels(
+                    method=request.method, status=500,
+                    **labels.values_for(response)
+                ).inc()
 
             return
 
@@ -613,6 +642,11 @@ class PrometheusMetrics(object):
         def decorator(f):
             @wraps(f)
             def func(*args, **kwargs):
+                if self.exclude_user_defaults and self.excluded_paths:
+                    # exclude based on default excludes
+                    if any(pattern.match(request.path) for pattern in self.excluded_paths):
+                        return f(*args, **kwargs)
+
                 if before:
                     metric = get_metric(None)
                     before(metric)
@@ -807,7 +841,8 @@ class PrometheusMetrics(object):
 
         gauge = Gauge(
             name, description, labelnames or tuple(),
-            registry=self.registry
+            registry=self.registry,
+            multiprocess_mode='max'
         )
 
         if labels:
@@ -824,6 +859,25 @@ class PrometheusMetrics(object):
         except NameError:
             return isinstance(value, str)  # python3
 
+    @staticmethod
+    def _not_yet_handled(tracking_key):
+        """
+        Check if the request has not handled some tracking yet,
+        and mark the request if this is the first time.
+        This is to avoid follow-up actions.
+
+        :param tracking_key: a key identifying a processing step
+        :return: True if this is the first time the request is
+          trying to handle this processing step
+        """
+
+        key = 'prom_' + tracking_key
+        if hasattr(request, key):
+            return False
+        else:
+            setattr(request, key, True)
+            return True
+
 
 class ConnexionPrometheusMetrics(PrometheusMetrics):
     """
@@ -905,4 +959,4 @@ class RESTfulPrometheusMetrics(PrometheusMetrics):
         return _make_response
 
 
-__version__ = '0.17.0'
+__version__ = '0.20.0'
